Global gSectors:TSector[] = Null
Global gBottomSectors:TList = Null

Type TSector
	Field x:Float = 0.0
	Field y:Float = 0.0
	Field width:Float = 0.0
	Field height:Float = 0.0
	Field entities:TList = CreateList() 'Bit of a misnomer. This stores only walls
	Field children:TList = CreateList()
	Field parent:TSector = Null
	Field marked:Byte = False
	Field tl:TVector = MakeVector(0.0,0.0)
	Field tr:TVector = MakeVector(0.0,0.0)
	Field bl:TVector = MakeVector(0.0,0.0)
	Field br:TVector = MakeVector(0.0,0.0)
	Field displayList:Int = -1
	Method SubDivide()
		ClearList children
		For Local i:Int = 0 To 3
			Local s:TSector = New TSector
			s.width = width / 2.0
			s.height = height / 2.0
			s.x = x
			s.y = y
			Select i
				Case 1
					s.x = x + s.width
				Case 2
					s.x = x + s.width
					s.y = y + s.height
				Case 3
					s.y = y + s.height
			End Select
			s.GetEntities()
			'Print CountList(s.entities)
			s.parent = Self
			s.MakeVectors()
			ListAddLast children, s
			If CountList(s.entities) > 8 And s.width > 4.0 And s.height > 4.0 Then
				'Print "BEFORE SUBDIVISION"
				s.SubDivide()
				'Print "AFTER SUBDIVISION"
			Else
				ListAddLast gBottomSectors, s
			EndIf
			
		Next
	End Method
	Method MakeVectors()
		tl.x=x
		tl.y=y
		tr.x=x+width
		tr.y=y		
		bl.x=x
		bl.y=y+height
		br.x=x+width
		br.y=y+height
	End Method
	Method GetEntities()
		ClearList entities
		For Local m:Int = 0 To Len(gMapWalls) - 1
			If gMapWalls[m] <> Null Then
				If gMapWalls[m].x+2.0 >= x And gMapWalls[m].x-2.0 <= x+width And gMapWalls[m].z+2.0 >= y And gMapWalls[m].z-2.0 <= y+height Then
					ListAddLast entities, gMapWalls[m]
				EndIf
			EndIf
		Next
		For Local e:TEntity = EachIn gEntities
			If Instr(e.class, "wall:") > 0 Or e.class = "secretwall" Then
				If e.x+e.radius >= x And e.x-e.radius <= x+width And e.z+e.radius >= y And e.z-e.radius <= y+height Then
					ListAddLast entities, e
				EndIf

			EndIf
		Next
	End Method
	Method Clear()
		ClearList entities
		For Local s:TSector = EachIn children
			s.Clear()
		Next
		ClearList children
	End Method
	Method ContainsInstance(e:TEntity)
		If e.x + e.radius >= x And e.x - e.radius <= x+width And e.z + e.radius >= y And e.z - e.radius <= y+height Then
			Return True
		EndIf
		Return False
	End Method
End Type

Function GetSectorsOnRay:TList(rs:TVector, rd:TVector, maxDist:Float)
	Local list:TList = CreateList()
	GetSectorsOnRay2(rs,rd, gSectors[0].parent, list, maxDist)
	Return list
End Function

Function GetSectorsOnRay2(rs:TVector, rd:TVector, sector:TSector, list:TList Var, maxDist:Float)
	For Local c:TSector = EachIn sector.children
		'Project all four corners onto ray's normal
		Local rn:TVector = MakeVector(-rd.y, rd.x)
		Local tlp:Float = DotVector(rn, c.tl)
		Local trp:Float = DotVector(rn, c.tr)
		Local blp:Float = DotVector(rn, c.bl)
		Local brp:Float = DotVector(rn, c.br)
		Local rp:Float = DotVector(rn, rs)
		Local minp:Float = Min(tlp, Min(trp, Min(blp,brp)))
		Local maxp:Float = Max(tlp, Max(trp, Max(blp,brp)))
		If rp > minp And rp < maxp Then
			Local plane:TPlane = MakePlane(rs, AddVector(rs,rn))
			Local tld:Float = PlaneDistance(plane, c.tl)
			Local trd:Float = PlaneDistance(plane, c.tr)
			Local bld:Float = PlaneDistance(plane, c.bl)
			Local brd:Float = PlaneDistance(plane, c.br)
			If (tld > 0 Or trd > 0 Or bld > 0 Or brd > 0) And (tld < maxDist Or trd < maxDist Or bld < maxDist Or brd < maxDist) Then
				If CountList(c.children) <= 0 Then
					'c.marked = True
					ListAddLast list, c
				Else
					GetSectorsOnRay2(rs, rd, c, list, maxDist)
				EndIf
			EndIf
		EndIf
	Next
End Function

Function GetSectorsInFrustrum:TList(rs:TVector, rd:TVector, fov:Float, maxDist:Float) 'This isn't very precise. Don't use it for now.
	Local range:Float = GetCos(90.0 - (fov / 2.0))
	Local list:TList = CreateList()
	GetSectorsInFrustrum2(rs, rd, range, gSectors[0].parent, list, maxDist)
	Return list
End Function

Function GetSectorsInFrustrum2(rs:TVector, rd:TVector, range:Float, sector:TSector, list:TList Var, maxDist:Float) 'Consider using circle-based tests instead
	For Local c:TSector = EachIn sector.children
		Local tlr:TVector = SubtractVector(c.tl, rs); tlr.Normalize()
		Local trr:TVector = SubtractVector(c.tr, rs); trr.Normalize()
		Local blr:TVector = SubtractVector(c.bl, rs); blr.Normalize()
		Local brr:TVector = SubtractVector(c.br, rs); brr.Normalize()
		Local tlp:Float = DotVector(rd, tlr)
		Local trp:Float = DotVector(rd, trr)
		Local blp:Float = DotVector(rd, blr)
		Local brp:Float = DotVector(rd, brr)
		If trp > range Or tlp > range Or blp > range Or brp > range Then
			Local rn:TVector = MakeVector(-rd.y, rd.x)
			Local plane:TPlane = MakePlane(rs, AddVector(rs,rn))
			Local tld:Float = 0.0; Local trd:Float = 0.0; Local bld:Float = 0.0; Local brd:Float = 0.0
			If tlp > range Then tld = PlaneDistance(plane, c.tl)
			If trp > range Then trd = PlaneDistance(plane, c.tr)
			If blp > range Then bld = PlaneDistance(plane, c.bl)
			If brp > range Then brd = PlaneDistance(plane, c.br)
			If (tld > 0 Or trd > 0 Or bld > 0 Or brd > 0) And (tld < maxDist Or trd < maxDist Or bld < maxDist Or brd < maxDist) Then
				If CountList(c.children) <= 0 Then
					c.marked = True
					ListAddLast list, c
				Else
					GetSectorsInFrustrum2(rs, rd, range, c, list, maxDist)
				EndIf
			EndIf
		EndIf
	Next

End Function

Function FindCurrentSector:TSector(x:Float, z:Float, dx:Float, dz:Float, radius:Float, sector:TSector)
	For Local c:TSector = EachIn sector.children
		If x+dx+radius >= c.x And x+dx-radius <= c.x + c.width And z+dz+radius >= c.y And z+dz-radius <= c.y + c.height Then
			If CountList(c.children) = 0 Then
				'If you're not fully inside the sector return the parent
				If x+dx-radius >= c.x And x+dx+radius < c.x + c.width And z+dz-radius >= c.y And z+dz+radius < c.y + c.height Then
					Return c
				Else
					Return c.parent
				EndIf
				Exit
			Else
				Return FindCurrentSector(x, z, dx, dz, radius, c)
			EndIf
		EndIf
	Next
End Function

Function DrawSector(s:TSector)
	Local offset:Float = 32.0
	Local offsetX:Float = 128.0
	glColor3f 1.0, 1.0, 1.0
	glBegin GL_LINE_LOOP
		glVertex2f s.x+offsetX,s.y + offset
		glVertex2f s.x+s.width+offsetX,s.y + offset
		glVertex2f s.x+s.width+offsetX,s.y+s.height + offset
		glVertex2f s.x+offsetX,s.y+s.height + offset
		glVertex2f s.x+offsetX,s.y + offset
	glEnd
	If s.marked = True Then
		glColor3f 1.0, 1.0, 0.0
		glBegin GL_LINE_LOOP
			glVertex2f s.x+offsetX + 1,s.y + offset + 1
			glVertex2f s.x+s.width+offsetX - 1,s.y + offset + 1
			glVertex2f s.x+s.width+offsetX - 1,s.y+s.height + offset - 1
			glVertex2f s.x+offsetX + 1,s.y+s.height + offset - 1
			glVertex2f s.x+offsetX + 1,s.y + offset + 1
		glEnd
		s.marked = False
	EndIf
	For Local ss:TSector = EachIn s.children
		DrawSector(ss)
	Next
End Function

